import pytest

from test_django_plugin.app.models import Author, Book


class TestFixturesInClass:
    @pytest.fixture(scope='module')
    def lem(self, module_transaction):
        with module_transaction('lem'):
            yield Author.objects.create(first_name='Stanislav', last_name='Lem')

    @pytest.fixture(scope='module')
    def fat(self, module_transaction):
        with module_transaction('fat'):
            yield Author.objects.create(first_name='Leo', last_name='Fat')

    @pytest.fixture(scope='module')
    def henry(self, module_transaction):
        with module_transaction('henry'):
            yield Author.objects.create(first_name='O', last_name='Henry')

    @pytest.fixture(scope='module')
    def solaris(self, lem, module_transaction):
        with module_transaction('solaris'):
            yield Book.objects.create(name='Solaris', year=1961, author=lem)

    def test_01(self):
        assert Book.objects.count() == 0

    def test_02(self, fat, lem):
        assert Book.objects.count() == 0
        assert Author.objects.count() == 2

    def test_03(self, fat, henry):
        assert Book.objects.count() == 0
        assert Author.objects.count() == 2

    def test_04(self, lem, henry):
        assert Book.objects.count() == 0
        assert Author.objects.count() == 2

    def test_10(self, solaris):
        assert Book.objects.count() == 1
        assert Author.objects.count() == 1

    def test_99(self):
        assert Book.objects.count() == 0



import json
import os
from contextlib import contextmanager
from io import BytesIO
from pprint import pformat

from django.http.multipartparser import LazyStream, Parser


class APIError(Exception):
    def __init__(self, *args, resp, expected):
        super().__init__(*args)
        self.resp = resp
        self.expected = expected


class RequestVerbose:
    def __init__(self, request, pretty_json=True):
        func = request.function
        self.module = func.__module__
        self.test_name = func.__name__

        self.pretty_json = pretty_json
        self.verbose = True
        self.records = []
        self._silent = False

    def prettify_multipart(self, s):
        chunks = {}
        for (_type, meta, obj) in Parser(LazyStream(BytesIO(s)), b'BoUnDaRyStRiNg'):
            if _type.strip() == 'file':
                chunks[meta['content-disposition'][1]['name'].decode('utf8')] = '<BINARY_FILE>'
        msg = 'Multipart Binary:\n{}'.format(chunks)
        return msg

    def prettify(self, s):
        if isinstance(s, bytes):
            try:
                s = s.decode('utf8')
            except UnicodeDecodeError:
                if self.dj_request['CONTENT_TYPE'].startswith('multipart/'):
                    return self.prettify_multipart(s)
                raise

        if isinstance(s, str):
            try:
                return json.loads(s)
            except:
                if len(s) < 100:
                    return s
                return '{} ...'.format(s[:100])
        return pformat(s)

    def log(self, resp):
        request = resp.request
        self.dj_request = request
        if 'wsgi.input' in request:
            cnt = request['wsgi.input']._FakePayload__content
            cnt.seek(0)
            payload = self.prettify(cnt.read())
        else:
            payload = None

        rec = {
            'method': request['REQUEST_METHOD'],
            'path': request['PATH_INFO'],
            'query': request['QUERY_STRING'],
            'payload': payload,
            'status_code': resp.status_code,
            'status_text': resp.status_text,
            'response_full': resp.serialize().decode('utf8'),
            'response_headers': resp.serialize_headers().decode('utf8'),
            'response': self.prettify(resp.content),
        }
        self.record(rec)

    def record(self, log_record):
        if self._silent:
            return
        self.records.append(log_record)
        self.verbose_print(log_record)

    def verbose_print(self, log_record):
        if not self.verbose:
            return
        print('\n{method} {path} {query}'.format(**log_record))
        if log_record['payload']:
            print('Payload:\n{payload}'.format(**log_record))
        print('\nResponse Code: {status_code} Content:\n{response}\n'.format(**log_record))

    def finish(self):
        pass

    @contextmanager
    def silent(self):
        self._silent = True
        yield
        self._silent = False


class RequestLogger(RequestVerbose):
    def __init__(self, request, pretty_json=True):
        marker = request.get_closest_marker('docme')
        if not marker or 'path' not in marker.kwargs:
            raise Exception('You should mark test with @pytest.mark.docme(path="PATH")')

        super().__init__(request, pretty_json)

        self.doc_path = marker.kwargs['path']
        self.verbose = request.config.getoption('verbose')

    def finish(self):
        path = os.environ.get('DOCS_ROOT', './.doc')
        if not os.path.exists(path):
            os.mkdir(path)
        fname = os.path.join(path, '{}.{}.json'.format(self.module, self.doc_path))
        with open(fname, 'w') as f:
            json.dump(self.records, f)


class RequestLoggerStub:
    def __call__(self, *args, **kwargs):
        pass

    def __getattr__(self, name):
        return self

    @contextmanager
    def silent(self):
        yield


request_logger = RequestLoggerStub()


class UserWrapper:
    """
    Wraps User object, add requests support
    """

    content_format = 'json'  # or multipart

    def __init__(self, user=None):
        from rest_framework.test import APIClient

        self.user = user
        self.client = APIClient()
        self.client.force_authenticate(self.user)
        if self.user:
            self.client.force_login(self.user)

    def __getattr__(self, name):
        if self.user is None:
            raise AttributeError
        else:
            return getattr(self.user, name)

    @contextmanager
    def set_format(self, format):
        old_format = self.content_format
        try:
            self.content_format = format
            yield
        finally:
            self.content_format = old_format

    def _method(self, method, url, data):
        resp = getattr(self.client, method)(url, data, format=self.content_format)
        request_logger.log(resp)
        return resp

    def post(self, url, data):
        return self._method('post', url, data)

    def patch(self, url, data):
        return self._method('patch', url, data)

    def put(self, url, data):
        return self._method('put', url, data)

    def get(self, url, data=None):
        return self._method('get', url, data or {})

    def delete(self, url, data=None):
        return self._method('delete', url, data or {})

    def json_method(self, method, *args, **kwargs):
        with self.set_format('json'):
            return getattr(self, method)(*args, **kwargs)

    def check_code(self, resp, expected):
        if resp.status_code != expected:
            msg = '{} {}'.format(resp.status_code, resp.content)[:1000]
            raise APIError(msg, resp=resp, expected=expected)

    def get_json(self, *args, expected=200, **kwargs):
        r = self.json_method('get', *args, **kwargs)
        self.check_code(r, expected)
        return r.json()

    def get_results(self, *args, **kwargs):
        return self.get_json(*args, **kwargs)['results']

    def post_json(self, *args, expected=201, **kwargs):
        r = self.json_method('post', *args, **kwargs)
        self.check_code(r, expected)
        return r.json()

    def patch_json(self, *args, expected=200, **kwargs):
        r = self.json_method('patch', *args, **kwargs)
        self.check_code(r, expected)
        return r.json()

    def put_json(self, *args, expected=200, **kwargs):
        r = self.json_method('put', *args, **kwargs)
        self.check_code(r, expected)
        return r.json()

    def delete_json(self, *args, expected=204, **kwargs):
        r = self.json_method('delete', *args, **kwargs)
        self.check_code(r, expected)
        assert not r.content, r.content


def create_user(username, groups=(), permissions=(), **kwargs):
    from django.contrib.auth import get_user_model
    from django.contrib.auth.models import Permission, Group
    from django.contrib.contenttypes.models import ContentType

    User = get_user_model()
    exists = User.objects.filter(username=username).first()
    if exists:
        user = exists
    else:
        email = kwargs.pop('email', '%s@gettipsi.com' % username)
        password = kwargs.pop('password', username)
        user = User.objects.create_user(username, email, password, **kwargs)
        for name in groups:
            group = Group.objects.filter(name=name).first()
            if not group:
                group = Group.objects.create(name=name)
            user.groups.add(group)

        for model, codename in permissions:
            content_type = ContentType.objects.get_for_model(model)
            try:
                p = Permission.objects.get(codename=codename, content_type=content_type)
            except Permission.DoesNotExist:
                p = Permission.objects.create(codename=codename, content_type=content_type)
            user.user_permissions.add(p)
        user.refresh_from_db()
    return UserWrapper(user)



import os
import time
from contextlib import suppress
import pytest
from tipsi_tools.unix import wait_socket, succ


def create_database():
    print('Create database')
    c = 0
    while c < 30:
        try:
            succ('echo create database plugin | psql -h localhost -p 40001 -U postgres')
            return
        except:
            c += 1
            time.sleep(1)


def pytest_configure(config):
    if config.getoption('docker_skip'):
        return

    with suppress(Exception):
        print('Stop old container if exists')
        os.system('docker stop testpostgres')
        time.sleep(1)
    try:
        print('Start docker')
        os.system('docker run -d -p 40001:5432 --name=testpostgres --rm=true postgres')
        time.sleep(5)
        wait_socket('localhost', 40001, timeout=15)
        create_database()
    except Exception as e:
        print('EXCEPTION: {}'.format(e))
        exit(1)


def pytest_unconfigure(config):
    if config.getoption('docker_skip'):
        return

    os.system('docker stop -t 2 testpostgres')


@pytest.fixture(autouse=True)
def autodatabase(module_transaction):
    pass



import pytest


@pytest.fixture
def invalid_fixture(module_transaction):
    with module_transaction('invalid_fixture'):
        raise Exception('FUFIX')


@pytest.mark.xfail
def test_invalid_fixture(invalid_fixture):
    pass



import inspect
from contextlib import contextmanager

import pytest
from _pytest import fixtures
from pytest_django.fixtures import SettingsWrapper as OriginalSettingsWrapper

_transactions_stack = []


def get_defs_by_name(vprint, request, name):
    fixturemanager = request.session._fixturemanager
    defs = fixturemanager.getfixturedefs(name, request.node.nodeid)
    # see test_transaction_class.py to find failed case
    # class fixtures have baseid = <module>::<Class>::()
    # but tests have nodeid = <module>
    # so class fixtures basically have more precise scope than function
    # thus we need to wider scope for such cases
    if not defs and name in fixturemanager._arg2fixturedefs:
        defs = fixturemanager._arg2fixturedefs[name]
        vprint(
            'Get fixtures without respecting factories: {} {}'.format(request.node.nodeid, defs),
            level=2)

    # We need this only to finish fixtures in _transactions_stack
    # if we have no defs here - there will be infinite loop in atomic_rollback:while below
    assert defs, 'Cannot find: {} Available: {}'.format(
        name, list(fixturemanager._arg2fixturedefs.keys()))
    vprint('Finish variants: {} For: {}'.format(defs, name), level=3)
    return defs


def finish_fixture(vprint, request, name):
    defs = get_defs_by_name(vprint, request, name)
    if defs:
        for d in reversed(defs):
            d.finish(request)


@contextmanager
def atomic_rollback(vprint, name, fixturename, fixturemanager):
    from django.db import transaction

    sid = transaction.savepoint()
    try:
        _transactions_stack.append(fixturename)
        vprint('transaction start: {} {}'.format(name, sid), level=2)
        yield
    except Exception:
        fname = _transactions_stack.pop()
        assert fname == fixturename
        raise
    else:
        if _transactions_stack:
            curr = _transactions_stack[-1]
            while curr and curr != fixturename:
                vprint('Stack: {} {}'.format(_transactions_stack, curr))
                finish_fixture(vprint, fixturemanager, curr)
                if _transactions_stack:
                    curr = _transactions_stack[-1]
                else:
                    curr = None
            if curr == fixturename:
                vprint('rollback {} {}'.format(name, sid), level=2)
                _transactions_stack.pop()
                transaction.savepoint_rollback(sid)


def get_atomic_rollback(request, vprint):
    def _inner(fixturename, *args, **kwargs):
        f = inspect.currentframe().f_back.f_code
        name = '{} at {}:{}'.format(f.co_name, f.co_filename, f.co_firstlineno)
        if args or kwargs:
            formatted = '{} / {} {}'.format(name, args, kwargs)
        else:
            formatted = name
        return atomic_rollback(vprint, formatted, fixturename, request)

    return _inner


@pytest.fixture(scope='module')
def module_transaction(request, vprint, django_db_blocker, django_db_setup):
    from django.db import connection, transaction

    django_db_blocker.unblock()
    with transaction.atomic():
        sid = transaction.savepoint()
        yield get_atomic_rollback(request, vprint)
        transaction.savepoint_rollback(sid)
    connection.close()


@pytest.fixture
def function_fixture(request, module_transaction):
    with module_transaction(request.fixturename):
        yield


@pytest.fixture(scope='module')
def module_fixture(request, module_transaction):
    with module_transaction(request.fixturename):
        yield


class SettingsWrapper(OriginalSettingsWrapper):
    """
    we need `_to_restore` to be object local to work with nesting
    see test_settings.py for test case
    """

    def __init__(self):
        self.__dict__['_to_restore'] = []
        assert id(SettingsWrapper._to_restore) != id(self._to_restore)


@pytest.fixture(scope='session')
def session_settings():
    wrapper = SettingsWrapper()
    yield wrapper
    wrapper.finalize()


@pytest.fixture(scope='module')
def module_settings(session_settings):
    wrapper = SettingsWrapper()
    yield wrapper
    wrapper.finalize()


@pytest.fixture(scope='function')
def function_settings(module_settings):
    wrapper = SettingsWrapper()
    yield wrapper
    wrapper.finalize()


def transaction_fx(scope):
    """
    Shortcut for fixtures which uses module_transaction
    >>>@pytest.fixture(scope='module')
    ...def some_fx(request, sub_fx, module_transaction):
    ...    with module_transaction(request.fixturename):
    ...        yield SomeModel.objects.create(sub=sub_fx)
    vs
    >>>@module_fx
    ...def some_fx(sub_fx):
    ...    return SomeModel.objects.create(sub=sub_fx)
    """

    def _decorator(func):
        arg_names = fixtures.getfuncargnames(func)

        def _inner(request):
            requested_fixtures = [request.getfixturevalue(n) for n in arg_names]
            module_transaction = request.getfixturevalue('module_transaction')
            with module_transaction(request.fixturename):
                yield func(*requested_fixtures)

        return pytest.fixture(scope=scope, name=func.__name__)(_inner)

    return _decorator


module_fx = transaction_fx('module')
function_fx = transaction_fx('function')

try:
    from pytest_django.fixtures import django_assert_num_queries
except ImportError:

    @pytest.fixture(scope='function')
    def django_assert_num_queries(pytestconfig):
        from django.db import connection
        from django.test.utils import CaptureQueriesContext

        @contextmanager
        def _assert_num_queries(num):
            with CaptureQueriesContext(connection) as context:
                yield
                if num != len(context):
                    msg = "Expected to perform %s queries but %s were done" % (num, len(context))
                    if pytestconfig.getoption('verbose') > 0:
                        sqls = (q['sql'] for q in context.captured_queries)
                        msg += '\n\nQueries:\n========\n\n%s' % '\n\n'.join(sqls)
                    else:
                        msg += " (add -v option to show queries)"
                    pytest.fail(msg)

        return _assert_num_queries




class TestSettings:
    def test_01(self, function_settings, session_settings):
        function_settings.CUSTOM_SETTINGS = 5
        from django.conf import settings as dj_settings
        assert dj_settings.CUSTOM_SETTINGS == 5
        assert dj_settings.BROKER_BACKEND == 'memory'

    def test_02(self):
        from django.conf import settings
        assert settings.BROKER_BACKEND == 'memory'



import pytest

from pytest_tipsi_django import client_fixtures
from pytest_tipsi_django.client_fixtures import UserWrapper
from pytest_tipsi_django.django_fixtures import (  # noqa
    session_settings, module_settings, function_settings,
    module_transaction, function_fixture, module_fixture,
    django_assert_num_queries
)


def setup_docs_logger(item):
    func = item.function
    if hasattr(func, 'docme') and item.config.getoption('write_docs'):
        client_fixtures.request_logger = client_fixtures.RequestLogger(item)
    elif item.config.getoption('verbose_request'):
        client_fixtures.request_logger = client_fixtures.RequestVerbose(item)


def pytest_addoption(parser):
    group = parser.getgroup('docme')
    group.addoption('--write-docs', action='store_true', default=False,
                    help='Write http requests for documentation')
    group.addoption('--verbose_request', action='store_true', default=False,
                    help='Print all requests to stdout')

def pytest_configure(config):
    options = ['write_docs', 'verbose_request']
    if all(config.getoption(opt) for opt in options):
        raise pytest.UsageError("Use only one option at the same time {}".format(options))


@pytest.fixture
def docs_logger():
    """
    Make requests without being logged:

    >>> def my_test(docs_logger):
    ...     with docs_logger.silent():
    ...        client.get_json(url)  # not logged
    ...     client.get_json(url)  # logged
    """
    yield client_fixtures.request_logger


def finish_docs_logger(item, nextitem):
    func = item.function
    if hasattr(func, 'docme'):
        client_fixtures.request_logger.finish()
        client_fixtures.request_logger = client_fixtures.RequestLoggerStub()


@pytest.fixture(autouse=True)  # noqa: F811
def auto_transaction(request):
    """
    It should be the last fixture, so we ensure that we've loaded all other required fixtures
    before open transaction.
    If we meet error here, try to modify `if` as `name not in (request.fixturename, OTHERNAME)`
    """
    if 'module_transaction' not in request.fixturenames:
        yield None
    else:
        # We cannot request module_transaction or it will be used for all our tests
        module_transaction = request.getfixturevalue('module_transaction')  # noqa: F811
        with module_transaction(request.fixturename):
            yield


@pytest.fixture(scope='session', autouse=True)  # noqa
def local_cache(session_settings, django_db_setup):
    session_settings.DEFAULT_FILE_STORAGE = 'django.core.files.storage.FileSystemStorage'
    session_settings.BROKER_BACKEND = 'memory'
    session_settings.CACHES = {
        'default': {
            'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
            'TIMEOUT': 60 * 15
        }
    }
    yield session_settings


@pytest.fixture
def anonymous_client(module_transaction):
    yield UserWrapper(None)


def pytest_runtest_setup(item):
    setup_docs_logger(item)


def pytest_runtest_teardown(item, nextitem):
    finish_docs_logger(item, nextitem)



from django.db import models

# Create your models here.


class Author(models.Model):
    first_name = models.CharField(max_length=255)
    last_name = models.CharField(max_length=255)


class Book(models.Model):
    name = models.CharField(max_length=255)
    year = models.IntegerField()
    author = models.ForeignKey(Author, on_delete=models.PROTECT)



from io import open

from setuptools import find_packages, setup

with open('pytest_tipsi_django/__init__.py', 'r') as f:
    for line in f:
        if line.startswith('__version__'):
            version = line.strip().split('=')[1].strip(' \'"')
            break
    else:
        version = '0.0.1'

with open('README.rst', 'r', encoding='utf-8') as f:
    readme = f.read()

REQUIRES = [
    'django',
    'pytest>=3.3.0',
    'pytest-django==3.2.*',
    'tipsi-tools>=1.7.0',
    'pytest-tipsi-testing>=1.3.0',
]

setup(
    name='pytest-tipsi-django',
    version=version,
    description='',
    long_description=readme,
    author='cybergrind',
    author_email='cybergrind@gmail.com',
    maintainer='cybergrind',
    maintainer_email='cybergrind@gmail.com',
    url='https://github.com/tipsi/pytest-tipsi-django',
    license='MIT',

    entry_points={
        'pytest11': ['pytest_tipsi_django = pytest_tipsi_django.plugin'],
    },

    keywords=[
        'pytest', 'pytest-django', 'fixtures', 'transactions', 'scopes',
    ],

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
    ],

    install_requires=REQUIRES,
    tests_require=['coverage', 'pytest'],

    packages=find_packages(),
)



import pytest

from pytest_tipsi_django.django_fixtures import module_fx, function_fx


from test_django_plugin.app.models import Author, Book


@module_fx
def lem():
    return Author.objects.create(first_name='Stanislav', last_name='Lem')


@function_fx
def solaris(lem):
    return Book.objects.create(name='Solaris', year=1961, author=lem)


def test_00():
    assert Author.objects.count() == 0
    assert Book.objects.count() == 0


author_id = None
book_id = None


def test_01(solaris, lem):
    global author_id, book_id
    assert Author.objects.count() == 1
    assert Book.objects.count() == 1
    author_id = id(lem)
    book_id = id(solaris)


def test_02(solaris, lem):
    assert id(lem) == author_id
    assert id(solaris) != book_id
    assert Author.objects.count() == 1
    assert Book.objects.count() == 1


def test_03():
    assert Author.objects.count() == 0
    assert Book.objects.count() == 0


def test_04(solaris):
    assert Author.objects.count() == 1
    assert Book.objects.count() == 1
